import { EventHandler } from '../../core/event-handler.js';
import { platform } from '../../core/platform.js';
import { XrInputSource } from './xr-input-source.js';

/**
 * @import { XrManager } from './xr-manager.js'
 */ /**
 * Provides access to input sources for WebXR.
 *
 * Input sources represent:
 *
 * - hand held controllers - and their optional capabilities: gamepad and vibration
 * - hands - with their individual joints
 * - transient sources - such as touch screen taps and voice commands
 *
 * @category XR
 */ class XrInput extends EventHandler {
    /** @private */ _onSessionStart() {
        var session = this.manager.session;
        session.addEventListener('inputsourceschange', this._onInputSourcesChangeEvt);
        session.addEventListener('select', (evt)=>{
            var inputSource = this._getByInputSource(evt.inputSource);
            inputSource.update(evt.frame);
            inputSource.fire('select', evt);
            this.fire('select', inputSource, evt);
        });
        session.addEventListener('selectstart', (evt)=>{
            var inputSource = this._getByInputSource(evt.inputSource);
            inputSource.update(evt.frame);
            inputSource._selecting = true;
            inputSource.fire('selectstart', evt);
            this.fire('selectstart', inputSource, evt);
        });
        session.addEventListener('selectend', (evt)=>{
            var inputSource = this._getByInputSource(evt.inputSource);
            inputSource.update(evt.frame);
            inputSource._selecting = false;
            inputSource.fire('selectend', evt);
            this.fire('selectend', inputSource, evt);
        });
        session.addEventListener('squeeze', (evt)=>{
            var inputSource = this._getByInputSource(evt.inputSource);
            inputSource.update(evt.frame);
            inputSource.fire('squeeze', evt);
            this.fire('squeeze', inputSource, evt);
        });
        session.addEventListener('squeezestart', (evt)=>{
            var inputSource = this._getByInputSource(evt.inputSource);
            inputSource.update(evt.frame);
            inputSource._squeezing = true;
            inputSource.fire('squeezestart', evt);
            this.fire('squeezestart', inputSource, evt);
        });
        session.addEventListener('squeezeend', (evt)=>{
            var inputSource = this._getByInputSource(evt.inputSource);
            inputSource.update(evt.frame);
            inputSource._squeezing = false;
            inputSource.fire('squeezeend', evt);
            this.fire('squeezeend', inputSource, evt);
        });
        // add input sources
        var inputSources = session.inputSources;
        for(var i = 0; i < inputSources.length; i++){
            this._addInputSource(inputSources[i]);
        }
    }
    /** @private */ _onSessionEnd() {
        var i = this._inputSources.length;
        while(i--){
            var inputSource = this._inputSources[i];
            this._inputSources.splice(i, 1);
            inputSource.fire('remove');
            this.fire('remove', inputSource);
        }
        var session = this.manager.session;
        session.removeEventListener('inputsourceschange', this._onInputSourcesChangeEvt);
    }
    /**
     * @param {XRInputSourcesChangeEvent} evt - WebXR input sources change event.
     * @private
     */ _onInputSourcesChange(evt) {
        // remove
        for(var i = 0; i < evt.removed.length; i++){
            this._removeInputSource(evt.removed[i]);
        }
        // add
        for(var i1 = 0; i1 < evt.added.length; i1++){
            this._addInputSource(evt.added[i1]);
        }
    }
    /**
     * @param {XRInputSource} xrInputSource - Input source to search for.
     * @returns {XrInputSource|null} The input source that matches the given WebXR input source or
     * null if no match is found.
     * @private
     */ _getByInputSource(xrInputSource) {
        for(var i = 0; i < this._inputSources.length; i++){
            if (this._inputSources[i].inputSource === xrInputSource) {
                return this._inputSources[i];
            }
        }
        return null;
    }
    /**
     * @param {XRInputSource} xrInputSource - Input source to add.
     * @private
     */ _addInputSource(xrInputSource) {
        if (this._getByInputSource(xrInputSource)) {
            return;
        }
        var inputSource = new XrInputSource(this.manager, xrInputSource);
        this._inputSources.push(inputSource);
        this.fire('add', inputSource);
    }
    /**
     * @param {XRInputSource} xrInputSource - Input source to remove.
     * @private
     */ _removeInputSource(xrInputSource) {
        for(var i = 0; i < this._inputSources.length; i++){
            if (this._inputSources[i].inputSource !== xrInputSource) {
                continue;
            }
            var inputSource = this._inputSources[i];
            this._inputSources.splice(i, 1);
            var h = inputSource.hitTestSources.length;
            while(h--){
                inputSource.hitTestSources[h].remove();
            }
            inputSource.fire('remove');
            this.fire('remove', inputSource);
            return;
        }
    }
    /**
     * @param {XRFrame} frame - XRFrame from requestAnimationFrame callback.
     * @ignore
     */ update(frame) {
        for(var i = 0; i < this._inputSources.length; i++){
            this._inputSources[i].update(frame);
        }
    }
    /**
     * List of active {@link XrInputSource} instances.
     *
     * @type {XrInputSource[]}
     */ get inputSources() {
        return this._inputSources;
    }
    /**
     * Create a new XrInput instance.
     *
     * @param {XrManager} manager - WebXR Manager.
     * @ignore
     */ constructor(manager){
        var _window_XRPose_prototype, _window_XRPose;
        super(), /**
     * @type {XrInputSource[]}
     * @private
     */ this._inputSources = [], /**
     * @type {boolean}
     * @ignore
     */ this.velocitiesSupported = false;
        this.manager = manager;
        this.velocitiesSupported = !!(platform.browser && ((_window_XRPose = window.XRPose) == null ? void 0 : (_window_XRPose_prototype = _window_XRPose.prototype) == null ? void 0 : _window_XRPose_prototype.hasOwnProperty('linearVelocity')));
        this._onInputSourcesChangeEvt = (evt)=>{
            this._onInputSourcesChange(evt);
        };
        this.manager.on('start', this._onSessionStart, this);
        this.manager.on('end', this._onSessionEnd, this);
    }
}
/**
     * Fired when a new {@link XrInputSource} is added to the list. The handler is passed the
     * {@link XrInputSource} that has been added.
     *
     * @event
     * @example
     * app.xr.input.on('add', (inputSource) => {
     *     // new input source is added
     * });
     */ XrInput.EVENT_ADD = 'add';
/**
     * Fired when an {@link XrInputSource} is removed from the list. The handler is passed the
     * {@link XrInputSource} that has been removed.
     *
     * @event
     * @example
     * app.xr.input.on('remove', (inputSource) => {
     *     // input source is removed
     * });
     */ XrInput.EVENT_REMOVE = 'remove';
/**
     * Fired when {@link XrInputSource} has triggered primary action. This could be pressing a
     * trigger button, or touching a screen. The handler is passed the {@link XrInputSource} that
     * triggered the `select` event and the XRInputSourceEvent event from the WebXR API.
     *
     * @event
     * @example
     * const ray = new pc.Ray();
     * app.xr.input.on('select', (inputSource, evt) => {
     *     ray.set(inputSource.getOrigin(), inputSource.getDirection());
     *     if (obj.intersectsRay(ray)) {
     *         // selected an object with input source
     *     }
     * });
     */ XrInput.EVENT_SELECT = 'select';
/**
     * Fired when {@link XrInputSource} has started to trigger primary action. The handler is
     * passed the {@link XrInputSource} that triggered the `selectstart` event and the
     * XRInputSourceEvent event from the WebXR API.
     *
     * @event
     * @example
     * app.xr.input.on('selectstart', (inputSource, evt) => {
     *     console.log('Select started');
     * });
     */ XrInput.EVENT_SELECTSTART = 'selectstart';
/**
     * Fired when {@link XrInputSource} has ended triggering primary action. The handler is passed
     * the {@link XrInputSource} that triggered the `selectend` event and the XRInputSourceEvent
     * event from the WebXR API.
     *
     * @event
     * @example
     * app.xr.input.on('selectend', (inputSource, evt) => {
     *     console.log('Select ended');
     * });
     */ XrInput.EVENT_SELECTEND = 'selectend';
/**
     * Fired when {@link XrInputSource} has triggered squeeze action. This is associated with
     * "grabbing" action on the controllers. The handler is passed the {@link XrInputSource} that
     * triggered the `squeeze` event and the XRInputSourceEvent event from the WebXR API.
     *
     * @event
     * @example
     * app.xr.input.on('squeeze', (inputSource, evt) => {
     *     console.log('Squeeze');
     * });
     */ XrInput.EVENT_SQUEEZE = 'squeeze';
/**
     * Fired when {@link XrInputSource} has started to trigger squeeze action. The handler is
     * passed the {@link XrInputSource} that triggered the `squeezestart` event and the
     * XRInputSourceEvent event from the WebXR API.
     *
     * @event
     * @example
     * app.xr.input.on('squeezestart', (inputSource, evt) => {
     *     if (obj.containsPoint(inputSource.getPosition())) {
     *         // grabbed an object
     *     }
     * });
     */ XrInput.EVENT_SQUEEZESTART = 'squeezestart';
/**
     * Fired when {@link XrInputSource} has ended triggering squeeze action. The handler is passed
     * the {@link XrInputSource} that triggered the `squeezeend` event and the XRInputSourceEvent
     * event from the WebXR API.
     *
     * @event
     * @example
     * app.xr.input.on('squeezeend', (inputSource, evt) => {
     *     console.log('Squeeze ended');
     * });
     */ XrInput.EVENT_SQUEEZEEND = 'squeezeend';

export { XrInput };
