import { Quat } from '../../core/math/quat.js';
import { Vec3 } from '../../core/math/vec3.js';
import { Vec4 } from '../../core/math/vec4.js';
import { GSplatData } from './gsplat-data.js';

var offscreen = null;
var ctx = null;
var readImageData = (imageBitmap)=>{
    if (!offscreen || offscreen.width !== imageBitmap.width || offscreen.height !== imageBitmap.height) {
        offscreen = new OffscreenCanvas(imageBitmap.width, imageBitmap.height);
        ctx = offscreen.getContext('2d');
        ctx.globalCompositeOperation = 'copy';
    }
    ctx.drawImage(imageBitmap, 0, 0);
    return ctx.getImageData(0, 0, imageBitmap.width, imageBitmap.height).data;
};
var SH_C0 = 0.28209479177387814;
class GSplatSogsIterator {
    constructor(data, p, r, s, c, sh){
        var lerp = (a, b, t)=>a * (1 - t) + b * t;
        // extract means for centers
        var { meta } = data;
        var { means, quats, scales, opacities, sh0, shN } = meta;
        var means_l_data = p && readImageData(data.means_l._levels[0]);
        var means_u_data = p && readImageData(data.means_u._levels[0]);
        var quats_data = r && readImageData(data.quats._levels[0]);
        var scales_data = s && readImageData(data.scales._levels[0]);
        var opacities_data = c && readImageData(data.opacities._levels[0]);
        var sh0_data = c && readImageData(data.sh0._levels[0]);
        var sh_labels_l_data = sh && readImageData(data.sh_labels_l._levels[0]);
        var sh_labels_u_data = sh && readImageData(data.sh_labels_u._levels[0]);
        var sh_centroids_data = sh && readImageData(data.sh_centroids._levels[0]);
        this.read = (i)=>{
            if (p) {
                var nx = lerp(means.mins[0], means.maxs[0], ((means_u_data[i * 4 + 0] << 8) + means_l_data[i * 4 + 0]) / 65535);
                var ny = lerp(means.mins[1], means.maxs[1], ((means_u_data[i * 4 + 1] << 8) + means_l_data[i * 4 + 1]) / 65535);
                var nz = lerp(means.mins[2], means.maxs[2], ((means_u_data[i * 4 + 2] << 8) + means_l_data[i * 4 + 2]) / 65535);
                p.x = Math.sign(nx) * (Math.exp(Math.abs(nx)) - 1);
                p.y = Math.sign(ny) * (Math.exp(Math.abs(ny)) - 1);
                p.z = Math.sign(nz) * (Math.exp(Math.abs(nz)) - 1);
            }
            if (r) {
                var qx = lerp(quats.mins[0], quats.maxs[0], quats_data[i * 4 + 0] / 255);
                var qy = lerp(quats.mins[1], quats.maxs[1], quats_data[i * 4 + 1] / 255);
                var qz = lerp(quats.mins[2], quats.maxs[2], quats_data[i * 4 + 2] / 255);
                var qw = Math.sqrt(Math.max(0, 1 - (qx * qx + qy * qy + qz * qz)));
                r.set(qy, qz, qw, qx);
            }
            if (s) {
                var sx = lerp(scales.mins[0], scales.maxs[0], scales_data[i * 4 + 0] / 255);
                var sy = lerp(scales.mins[1], scales.maxs[1], scales_data[i * 4 + 1] / 255);
                var sz = lerp(scales.mins[2], scales.maxs[2], scales_data[i * 4 + 2] / 255);
                s.set(sx, sy, sz);
            }
            if (c) {
                var _$r = lerp(sh0.mins[0], sh0.maxs[0], sh0_data[i * 4 + 0] / 255);
                var g = lerp(sh0.mins[1], sh0.maxs[1], sh0_data[i * 4 + 1] / 255);
                var b = lerp(sh0.mins[2], sh0.maxs[2], sh0_data[i * 4 + 2] / 255);
                var a = lerp(opacities.mins[0], opacities.maxs[0], opacities_data[i * 4 + 0] / 255);
                c.set(0.5 + _$r * SH_C0, 0.5 + g * SH_C0, 0.5 + b * SH_C0, 1.0 / (1.0 + Math.exp(a * -1)));
            }
            if (sh) {
                var n = sh_labels_l_data[i * 4 + 0] + (sh_labels_u_data[i * 4 + 0] << 8);
                var u = n % 64 * 15;
                var v = Math.floor(n / 64);
                for(var j = 0; j < 3; ++j){
                    for(var k = 0; k < 15; ++k){
                        sh[j * 15 + k] = lerp(shN.mins, shN.maxs, sh_centroids_data[(u + k) * 4 + j + v * data.sh_centroids.width * 4] / 255);
                    }
                }
            }
        };
    }
}
class GSplatSogsData {
    createIter(p, r, s, c, sh) {
        return new GSplatSogsIterator(this, p, r, s, c, sh);
    }
    calcAabb(result) {
        var { mins, maxs } = this.meta.means;
        var map = (v)=>Math.sign(v) * (Math.exp(Math.abs(v)) - 1);
        result.center.set((map(mins[0]) + map(maxs[0])) * 0.5, (map(mins[1]) + map(maxs[1])) * 0.5, (map(mins[2]) + map(maxs[2])) * 0.5);
        result.halfExtents.set((map(maxs[0]) - map(mins[0])) * 0.5, (map(maxs[1]) - map(mins[1])) * 0.5, (map(maxs[2]) - map(mins[2])) * 0.5);
    }
    getCenters(result) {
        var p = new Vec3();
        var iter = this.createIter(p);
        for(var i = 0; i < this.numSplats; i++){
            iter.read(i);
            result[i * 3 + 0] = p.x;
            result[i * 3 + 1] = p.y;
            result[i * 3 + 2] = p.z;
        }
    }
    calcFocalPoint(result, pred) {
        result.set(0, 0, 0);
    }
    get isSogs() {
        return true;
    }
    decompress() {
        var members = [
            'x',
            'y',
            'z',
            'f_dc_0',
            'f_dc_1',
            'f_dc_2',
            'opacity',
            'scale_0',
            'scale_1',
            'scale_2',
            'rot_0',
            'rot_1',
            'rot_2',
            'rot_3'
        ];
        var { shBands } = this;
        // allocate spherical harmonics data
        if (shBands > 0) {
            var shMembers = [];
            for(var i = 0; i < 45; ++i){
                shMembers.push("f_rest_" + i);
            }
            members.splice(members.indexOf('f_dc_0') + 1, 0, ...shMembers);
        }
        // allocate uncompressed data
        var data = {};
        members.forEach((name)=>{
            data[name] = new Float32Array(this.numSplats);
        });
        var p = new Vec3();
        var r = new Quat();
        var s = new Vec3();
        var c = new Vec4();
        var sh = shBands > 0 ? new Float32Array(45) : null;
        var iter = this.createIter(p, r, s, c, sh);
        for(var i1 = 0; i1 < this.numSplats; ++i1){
            iter.read(i1);
            data.x[i1] = p.x;
            data.y[i1] = p.y;
            data.z[i1] = p.z;
            data.rot_1[i1] = r.x;
            data.rot_2[i1] = r.y;
            data.rot_3[i1] = r.z;
            data.rot_0[i1] = r.w;
            data.scale_0[i1] = s.x;
            data.scale_1[i1] = s.y;
            data.scale_2[i1] = s.z;
            data.f_dc_0[i1] = (c.x - 0.5) / SH_C0;
            data.f_dc_1[i1] = (c.y - 0.5) / SH_C0;
            data.f_dc_2[i1] = (c.z - 0.5) / SH_C0;
            // convert opacity to log sigmoid taking into account infinities at 0 and 1
            data.opacity[i1] = c.w <= 0 ? -40 : c.w >= 1 ? 40 : -Math.log(1 / c.w - 1);
            if (sh) {
                for(var c1 = 0; c1 < 45; ++c1){
                    data["f_rest_" + c1][i1] = sh[c1];
                }
            }
        }
        return new GSplatData([
            {
                name: 'vertex',
                count: this.numSplats,
                properties: members.map((name)=>{
                    return {
                        name: name,
                        type: 'float',
                        byteSize: 4,
                        storage: data[name]
                    };
                })
            }
        ]);
    }
}

export { GSplatSogsData };
