import { Vec3 } from '../../core/math/vec3.js';
import { FILTER_NEAREST, TEXTURETYPE_DEFAULT, ADDRESS_CLAMP_TO_EDGE, PIXELFORMAT_RGBA32F } from '../../platform/graphics/constants.js';
import { FloatPacking } from '../../core/math/float-packing.js';
import { LIGHTTYPE_SPOT, LIGHTSHAPE_PUNCTUAL, LIGHT_COLOR_DIVIDER, LIGHTSHAPE_SPHERE, LIGHTSHAPE_DISK, LIGHTSHAPE_RECT } from '../constants.js';
import { Texture } from '../../platform/graphics/texture.js';
import { LightCamera } from '../renderer/light-camera.js';
import { shaderChunks } from '../shader-lib/chunks/chunks.js';
import { shaderChunksWGSL } from '../shader-lib/chunks-wgsl/chunks-wgsl.js';

var tempVec3 = new Vec3();
var tempAreaLightSizes = new Float32Array(6);
var areaHalfAxisWidth = new Vec3(-0.5, 0, 0);
var areaHalfAxisHeight = new Vec3(0, 0, 0.5);
// format of the float texture data
var TextureIndexFloat = {
    POSITION_RANGE: 0,
    DIRECTION_FLAGS: 1,
    COLOR_ANGLES_BIAS: 2,
    PROJ_MAT_0: 3,
    ATLAS_VIEWPORT: 3,
    PROJ_MAT_1: 4,
    PROJ_MAT_2: 5,
    PROJ_MAT_3: 6,
    AREA_DATA_WIDTH: 7,
    AREA_DATA_HEIGHT: 8,
    // leave last
    COUNT: 9
};
// enums supplied to the shader as inject-defines
var enums = {
    '{LIGHTSHAPE_PUNCTUAL}': "" + LIGHTSHAPE_PUNCTUAL + "u",
    '{LIGHTSHAPE_RECT}': "" + LIGHTSHAPE_RECT + "u",
    '{LIGHTSHAPE_DISK}': "" + LIGHTSHAPE_DISK + "u",
    '{LIGHTSHAPE_SPHERE}': "" + LIGHTSHAPE_SPHERE + "u",
    '{LIGHT_COLOR_DIVIDER}': "" + LIGHT_COLOR_DIVIDER + ".0"
};
// converts object with properties to a list of these as an example: "#define CLUSTER_TEXTURE_8_BLAH 1"
var buildShaderDefines = (object, prefix)=>{
    return Object.keys(object).map((key)=>"#define " + prefix + key + " " + object[key]).join('\n');
};
// create a shader chunk with defines for the light buffer textures
shaderChunks.lightBufferDefinesPS = shaderChunksWGSL.lightBufferDefinesPS = "\n\n    " + buildShaderDefines(TextureIndexFloat, 'CLUSTER_TEXTURE_') + "\n    " + buildShaderDefines(enums, '') + "\n";
// A class used by clustered lighting, responsible for encoding light properties into textures for the use on the GPU
class LightsBuffer {
    destroy() {
        var // release texture
        _this_lightsTexture;
        (_this_lightsTexture = this.lightsTexture) == null ? void 0 : _this_lightsTexture.destroy();
        this.lightsTexture = null;
    }
    createTexture(device, width, height, format, name) {
        var tex = new Texture(device, {
            name: name,
            width: width,
            height: height,
            mipmaps: false,
            format: format,
            addressU: ADDRESS_CLAMP_TO_EDGE,
            addressV: ADDRESS_CLAMP_TO_EDGE,
            type: TEXTURETYPE_DEFAULT,
            magFilter: FILTER_NEAREST,
            minFilter: FILTER_NEAREST,
            anisotropy: 1
        });
        return tex;
    }
    setBounds(min, delta) {
        this.boundsMin.copy(min);
        this.boundsDelta.copy(delta);
    }
    uploadTextures() {
        this.lightsTexture.lock().set(this.lightsFloat);
        this.lightsTexture.unlock();
    }
    updateUniforms() {
        // texture
        this._lightsTextureId.setValue(this.lightsTexture);
    }
    getSpotDirection(direction, spot) {
        // Spots shine down the negative Y axis
        var mat = spot._node.getWorldTransform();
        mat.getY(direction).mulScalar(-1);
        direction.normalize();
    }
    // half sizes of area light in world space, returned as an array of 6 floats
    getLightAreaSizes(light) {
        var mat = light._node.getWorldTransform();
        mat.transformVector(areaHalfAxisWidth, tempVec3);
        tempAreaLightSizes[0] = tempVec3.x;
        tempAreaLightSizes[1] = tempVec3.y;
        tempAreaLightSizes[2] = tempVec3.z;
        mat.transformVector(areaHalfAxisHeight, tempVec3);
        tempAreaLightSizes[3] = tempVec3.x;
        tempAreaLightSizes[4] = tempVec3.y;
        tempAreaLightSizes[5] = tempVec3.z;
        return tempAreaLightSizes;
    }
    // fill up both float and 8bit texture data with light properties
    addLightData(light, lightIndex) {
        var isSpot = light._type === LIGHTTYPE_SPOT;
        var hasAtlasViewport = light.atlasViewportAllocated; // if the light does not have viewport, it does not fit to the atlas
        var isCookie = this.cookiesEnabled && !!light._cookie && hasAtlasViewport;
        var isArea = this.areaLightsEnabled && light.shape !== LIGHTSHAPE_PUNCTUAL;
        var castShadows = this.shadowsEnabled && light.castShadows && hasAtlasViewport;
        var pos = light._node.getPosition();
        var lightProjectionMatrix = null; // light projection matrix - used for shadow map and cookie of spot light
        var atlasViewport = null; // atlas viewport info - used for shadow map and cookie of omni light
        if (isSpot) {
            if (castShadows) {
                var lightRenderData = light.getRenderData(null, 0);
                lightProjectionMatrix = lightRenderData.shadowMatrix;
            } else if (isCookie) {
                lightProjectionMatrix = LightCamera.evalSpotCookieMatrix(light);
            }
        } else {
            if (castShadows || isCookie) {
                atlasViewport = light.atlasViewport;
            }
        }
        var dataFloat = this.lightsFloat;
        var dataUint = this.lightsUint;
        var dataFloatStart = lightIndex * this.lightsTexture.width * 4;
        // pos and range
        dataFloat[dataFloatStart + 4 * TextureIndexFloat.POSITION_RANGE + 0] = pos.x;
        dataFloat[dataFloatStart + 4 * TextureIndexFloat.POSITION_RANGE + 1] = pos.y;
        dataFloat[dataFloatStart + 4 * TextureIndexFloat.POSITION_RANGE + 2] = pos.z;
        dataFloat[dataFloatStart + 4 * TextureIndexFloat.POSITION_RANGE + 3] = light.attenuationEnd;
        // color, spot angles, biases
        var clusteredData = light.clusteredData;
        dataUint[dataFloatStart + 4 * TextureIndexFloat.COLOR_ANGLES_BIAS + 0] = clusteredData[0];
        dataUint[dataFloatStart + 4 * TextureIndexFloat.COLOR_ANGLES_BIAS + 1] = clusteredData[1];
        dataUint[dataFloatStart + 4 * TextureIndexFloat.COLOR_ANGLES_BIAS + 2] = clusteredData[2];
        // biases (those are non-constant values, needs simplification)
        if (light.castShadows) {
            var lightRenderData1 = light.getRenderData(null, 0);
            var biases = light._getUniformBiasValues(lightRenderData1);
            // store them in 32bits as half floats
            var biasHalf = FloatPacking.float2Half(biases.bias);
            var normalBiasHalf = FloatPacking.float2Half(biases.normalBias);
            dataUint[dataFloatStart + 4 * TextureIndexFloat.COLOR_ANGLES_BIAS + 3] = biasHalf | normalBiasHalf << 16;
        }
        // spot direction
        if (isSpot) {
            this.getSpotDirection(tempVec3, light);
            dataFloat[dataFloatStart + 4 * TextureIndexFloat.DIRECTION_FLAGS + 0] = tempVec3.x;
            dataFloat[dataFloatStart + 4 * TextureIndexFloat.DIRECTION_FLAGS + 1] = tempVec3.y;
            dataFloat[dataFloatStart + 4 * TextureIndexFloat.DIRECTION_FLAGS + 2] = tempVec3.z;
        // here we have unused float
        }
        // flags
        dataUint[dataFloatStart + 4 * TextureIndexFloat.DIRECTION_FLAGS + 3] = light.getClusteredFlags(castShadows, isCookie);
        // light projection matrix
        if (lightProjectionMatrix) {
            var matData = lightProjectionMatrix.data;
            for(var m = 0; m < 16; m++){
                dataFloat[dataFloatStart + 4 * TextureIndexFloat.PROJ_MAT_0 + m] = matData[m];
            }
        }
        if (atlasViewport) {
            dataFloat[dataFloatStart + 4 * TextureIndexFloat.ATLAS_VIEWPORT + 0] = atlasViewport.x;
            dataFloat[dataFloatStart + 4 * TextureIndexFloat.ATLAS_VIEWPORT + 1] = atlasViewport.y;
            dataFloat[dataFloatStart + 4 * TextureIndexFloat.ATLAS_VIEWPORT + 2] = atlasViewport.z / 3; // size of a face slot (3x3 grid)
        }
        // area light sizes
        if (isArea) {
            var areaSizes = this.getLightAreaSizes(light);
            dataFloat[dataFloatStart + 4 * TextureIndexFloat.AREA_DATA_WIDTH + 0] = areaSizes[0];
            dataFloat[dataFloatStart + 4 * TextureIndexFloat.AREA_DATA_WIDTH + 1] = areaSizes[1];
            dataFloat[dataFloatStart + 4 * TextureIndexFloat.AREA_DATA_WIDTH + 2] = areaSizes[2];
            dataFloat[dataFloatStart + 4 * TextureIndexFloat.AREA_DATA_HEIGHT + 0] = areaSizes[3];
            dataFloat[dataFloatStart + 4 * TextureIndexFloat.AREA_DATA_HEIGHT + 1] = areaSizes[4];
            dataFloat[dataFloatStart + 4 * TextureIndexFloat.AREA_DATA_HEIGHT + 2] = areaSizes[5];
        }
    }
    constructor(device){
        this.areaLightsEnabled = false;
        this.device = device;
        // features
        this.cookiesEnabled = false;
        this.shadowsEnabled = false;
        this.areaLightsEnabled = false;
        // using 8 bit index so this is maximum supported number of lights
        this.maxLights = 255;
        // float texture
        var pixelsPerLightFloat = TextureIndexFloat.COUNT;
        this.lightsFloat = new Float32Array(4 * pixelsPerLightFloat * this.maxLights);
        this.lightsUint = new Uint32Array(this.lightsFloat.buffer);
        this.lightsTexture = this.createTexture(this.device, pixelsPerLightFloat, this.maxLights, PIXELFORMAT_RGBA32F, 'LightsTexture');
        this._lightsTextureId = this.device.scope.resolve('lightsTexture');
        // compression ranges
        this.invMaxColorValue = 0;
        this.invMaxAttenuation = 0;
        this.boundsMin = new Vec3();
        this.boundsDelta = new Vec3();
    }
}

export { LightsBuffer };
