import { Debug } from '../../../core/debug.js';
import { SHADER_FORWARD, SPRITE_RENDERMODE_SLICED, SPRITE_RENDERMODE_TILED, ditherNames, FRESNEL_SCHLICK, BLEND_NONE, DITHER_NONE } from '../../constants.js';
import { ShaderPass } from '../../shader-pass.js';
import { LitShader } from './lit-shader.js';
import { ChunkUtils } from '../chunk-utils.js';
import { StandardMaterialOptions } from '../../materials/standard-material-options.js';
import { LitOptionsUtils } from './lit-options-utils.js';
import { ShaderGenerator } from './shader-generator.js';
import { ShaderUtils } from '../../../platform/graphics/shader-utils.js';
import { SHADERLANGUAGE_GLSL, SHADERTAG_MATERIAL } from '../../../platform/graphics/constants.js';

function _extends() {
    _extends = Object.assign || function(target) {
        for(var i = 1; i < arguments.length; i++){
            var source = arguments[i];
            for(var key in source){
                if (Object.prototype.hasOwnProperty.call(source, key)) {
                    target[key] = source[key];
                }
            }
        }
        return target;
    };
    return _extends.apply(this, arguments);
}
/**
 * @import { GraphicsDevice } from '../../../platform/graphics/graphics-device.js'
 */ var _matTex2D = [];
var buildPropertiesList = (options)=>{
    return Object.keys(options).filter((key)=>key !== 'litOptions').sort();
};
class ShaderGeneratorStandard extends ShaderGenerator {
    generateKey(options) {
        var props;
        if (options === this.optionsContextMin) {
            if (!this.propsMin) this.propsMin = buildPropertiesList(options);
            props = this.propsMin;
        } else if (options === this.optionsContext) {
            if (!this.props) this.props = buildPropertiesList(options);
            props = this.props;
        } else {
            props = buildPropertiesList(options);
        }
        var definesHash = ShaderGenerator.definesHash(options.defines);
        var key = "standard:\n" + definesHash + "\n" + props.map((prop)=>prop + options[prop]).join('\n') + LitOptionsUtils.generateKey(options.litOptions);
        return key;
    }
    /**
     * Get the code with which to to replace '*_TEXTURE_UV' in the map shader functions.
     *
     * @param {string} transformPropName - Name of the transform id in the options block. Usually "basenameTransform".
     * @param {string} uVPropName - Name of the UV channel in the options block. Usually "basenameUv".
     * @param {object} options - The options passed into createShaderDefinition.
     * @returns {string} The code used to replace '*_TEXTURE_UV' in the shader code.
     * @private
     */ _getUvSourceExpression(transformPropName, uVPropName, options) {
        var transformId = options[transformPropName];
        var uvChannel = options[uVPropName];
        var isMainPass = options.litOptions.pass === SHADER_FORWARD;
        var expression;
        if (isMainPass && options.litOptions.nineSlicedMode === SPRITE_RENDERMODE_SLICED) {
            expression = 'nineSlicedUv';
        } else if (isMainPass && options.litOptions.nineSlicedMode === SPRITE_RENDERMODE_TILED) {
            expression = 'nineSlicedUv';
        } else {
            if (transformId === 0) {
                expression = "vUv" + uvChannel;
            } else {
                // note: different capitalization!
                expression = "vUV" + uvChannel + "_" + transformId;
            }
            // if heightmap is enabled all maps except the heightmap are offset
            if (options.heightMap && transformPropName !== 'heightMapTransform') {
                expression += ' + dUvOffset';
            }
        }
        return expression;
    }
    _validateMapChunk(propName, chunkName, chunks) {
        Debug.call(()=>{
            var code = chunks[chunkName];
            var requiredChangeStrings = [];
            // Helper function to add a formatted change string if the old syntax is found
            var trackChange = (oldSyntax, newSyntax)=>{
                if (code.includes(oldSyntax)) {
                    requiredChangeStrings.push("  " + oldSyntax + " -> " + newSyntax);
                }
            };
            // inject defines (with curly braces)
            [
                [
                    '$UV',
                    "{STD_" + propName + "_TEXTURE_UV}"
                ],
                [
                    '$CH',
                    "{STD_" + propName + "_TEXTURE_CHANNEL}"
                ],
                [
                    '$SAMPLER',
                    "{STD_" + propName + "_TEXTURE_NAME}"
                ],
                [
                    '$DECODE',
                    "{STD_" + propName + "_TEXTURE_DECODE}"
                ],
                [
                    '$VC',
                    "{STD_" + propName + "_VERTEX_CHANNEL}"
                ],
                [
                    '$DETAILMODE',
                    "{STD_" + propName + "_DETAILMODE}"
                ],
                [
                    'unpackNormal(',
                    "{STD_" + propName + "_TEXTURE_DECODE}("
                ]
            ].forEach((param)=>{
                var [oldSyntax, newSyntax] = param;
                return trackChange(oldSyntax, newSyntax);
            });
            // defines
            [
                [
                    'MAPFLOAT',
                    "STD_" + propName + "_CONSTANT"
                ],
                [
                    'MAPCOLOR',
                    "STD_" + propName + "_CONSTANT"
                ],
                [
                    'MAPVERTEX',
                    "STD_" + propName + "_VERTEX"
                ],
                [
                    'MAPTEXTURE',
                    "STD_" + propName + "_TEXTURE"
                ],
                [
                    'MAPINVERT',
                    "STD_" + propName + "_INVERT"
                ]
            ].forEach((param)=>{
                var [oldSyntax, newSyntax] = param;
                return trackChange(oldSyntax, newSyntax);
            });
            // custom handling
            if (code.includes('$texture2DSAMPLE')) {
                trackChange('$texture2DSAMPLE', '(Macro no longer supported - remove/refactor)');
            }
            if (requiredChangeStrings.length > 0) {
                Debug.errorOnce("Shader chunk " + chunkName + " is in no longer compatible format. Please make these replacements to bring it to the current version:\n" + requiredChangeStrings.join('\n'), {
                    code: code
                });
            }
        });
    }
    /**
     * Add shader defines for a texture map.
     *
     * @param {Map<string, string>} fDefines - The fragment defines.
     * @param {string} propName - The base name of the map: diffuse | emissive | opacity | light | height | metalness | specular | gloss | ao.
     * @param {string} chunkName - The name of the chunk to use. Usually "basenamePS".
     * @param {object} options - The options passed into to createShaderDefinition.
     * @param {object} chunks - The set of shader chunks to choose from.
     * @param {object} mapping - The mapping between chunk and sampler
     * @param {string|null} encoding - The texture's encoding
     * @private
     */ _addMapDefines(fDefines, propName, chunkName, options, chunks, mapping, encoding) {
        if (encoding === void 0) encoding = null;
        var mapPropName = "" + propName + "Map";
        var propNameCaps = propName.toUpperCase();
        var uVPropName = "" + mapPropName + "Uv";
        var identifierPropName = "" + mapPropName + "Identifier";
        var transformPropName = "" + mapPropName + "Transform";
        var channelPropName = "" + mapPropName + "Channel";
        var vertexColorChannelPropName = "" + propName + "VertexColorChannel";
        var tintPropName = "" + propName + "Tint";
        var vertexColorPropName = "" + propName + "VertexColor";
        var detailModePropName = "" + propName + "Mode";
        var invertName = "" + propName + "Invert";
        var tintOption = options[tintPropName];
        var vertexColorOption = options[vertexColorPropName];
        var textureOption = options[mapPropName];
        var textureIdentifier = options[identifierPropName];
        var detailModeOption = options[detailModePropName];
        var chunkCode = chunks[chunkName];
        // log errors if the chunk format is deprecated (format changed in engine 2.7)
        Debug.call(()=>{
            if (chunkCode) {
                this._validateMapChunk(propNameCaps, chunkName, chunks);
            }
        });
        if (textureOption) {
            fDefines.set("STD_" + propNameCaps + "_TEXTURE", '');
            var uv = this._getUvSourceExpression(transformPropName, uVPropName, options);
            // chunk injection defines
            fDefines.set("{STD_" + propNameCaps + "_TEXTURE_UV}", uv);
            fDefines.set("{STD_" + propNameCaps + "_TEXTURE_CHANNEL}", options[channelPropName]);
            // texture sampler define
            var textureId = "{STD_" + propNameCaps + "_TEXTURE_NAME}";
            if (chunkCode.includes(textureId)) {
                var samplerName = "texture_" + mapPropName;
                var alias = mapping[textureIdentifier];
                if (alias) {
                    samplerName = alias;
                } else {
                    mapping[textureIdentifier] = samplerName;
                    // texture is not aliased to existing texture, create a new one
                    fDefines.set("STD_" + propNameCaps + "_TEXTURE_ALLOCATE", '');
                }
                fDefines.set(textureId, samplerName);
            }
            if (encoding) {
                // decode function, ignored for alpha channel
                var textureDecode = options[channelPropName] === 'aaa' ? 'passThrough' : ChunkUtils.decodeFunc(encoding);
                fDefines.set("{STD_" + propNameCaps + "_TEXTURE_DECODE}", textureDecode);
            }
        }
        if (vertexColorOption) {
            fDefines.set("STD_" + propNameCaps + "_VERTEX", '');
            fDefines.set("{STD_" + propNameCaps + "_VERTEX_CHANNEL}", options[vertexColorChannelPropName]);
        }
        if (detailModeOption) {
            fDefines.set("{STD_" + propNameCaps + "_DETAILMODE}", detailModeOption);
        }
        if (tintOption) {
            fDefines.set("STD_" + propNameCaps + "_CONSTANT", '');
        }
        if (!!options[invertName]) {
            fDefines.set("STD_" + propNameCaps + "_INVERT", '');
        }
    }
    _correctChannel(p, chan, _matTex2D) {
        if (_matTex2D[p] > 0) {
            if (_matTex2D[p] < chan.length) {
                return chan.substring(0, _matTex2D[p]);
            } else if (_matTex2D[p] > chan.length) {
                var str = chan;
                var chr = str.charAt(str.length - 1);
                var addLen = _matTex2D[p] - str.length;
                for(var i = 0; i < addLen; i++)str += chr;
                return str;
            }
            return chan;
        }
    }
    createVertexShader(litShader, options) {
        var useUv = [];
        var useUnmodifiedUv = [];
        var mapTransforms = [];
        var maxUvSets = 2;
        for(var p in _matTex2D){
            var mapName = "" + p + "Map";
            if (options["" + p + "VertexColor"]) {
                var colorChannelName = "" + p + "VertexColorChannel";
                options[colorChannelName] = this._correctChannel(p, options[colorChannelName], _matTex2D);
            }
            if (options[mapName]) {
                var channelName = "" + mapName + "Channel";
                var transformName = "" + mapName + "Transform";
                var uvName = "" + mapName + "Uv";
                options[uvName] = Math.min(options[uvName], maxUvSets - 1);
                options[channelName] = this._correctChannel(p, options[channelName], _matTex2D);
                var uvSet = options[uvName];
                useUv[uvSet] = true;
                useUnmodifiedUv[uvSet] = useUnmodifiedUv[uvSet] || options[mapName] && !options[transformName];
                // create map transforms
                if (options[transformName]) {
                    mapTransforms.push({
                        name: p,
                        id: options[transformName],
                        uv: options[uvName]
                    });
                }
            }
        }
        if (options.forceUv1) {
            useUv[1] = true;
            useUnmodifiedUv[1] = useUnmodifiedUv[1] !== undefined ? useUnmodifiedUv[1] : true;
        }
        litShader.generateVertexShader(useUv, useUnmodifiedUv, mapTransforms);
    }
    /**
     * @param {StandardMaterialOptions} options - The create options.
     * @param {Map<string, string>} fDefines - The fragment defines.
     * @param {ShaderPass} shaderPassInfo - The shader pass info.
     */ prepareFragmentDefines(options, fDefines, shaderPassInfo) {
        var fDefineSet = (condition, name, value)=>{
            if (value === void 0) value = '';
            if (condition) {
                fDefines.set(name, value);
            }
        };
        fDefineSet(options.lightMap, 'STD_LIGHTMAP', '');
        fDefineSet(options.lightVertexColor, 'STD_LIGHT_VERTEX_COLOR', '');
        fDefineSet(options.dirLightMap && options.litOptions.useSpecular, 'STD_LIGHTMAP_DIR', '');
        fDefineSet(options.heightMap, 'STD_HEIGHT_MAP', '');
        fDefineSet(options.useSpecularColor, 'STD_SPECULAR_COLOR', '');
        fDefineSet(options.aoMap || options.aoVertexColor || options.useAO, 'STD_AO', '');
        fDefineSet(true, 'STD_OPACITY_DITHER', ditherNames[shaderPassInfo.isForward ? options.litOptions.opacityDither : options.litOptions.opacityShadowDither]);
    }
    /**
     * @param {GraphicsDevice} device - The graphics device.
     * @param {StandardMaterialOptions} options - The create options.
     * @returns {object} Returns the created shader definition.
     */ createShaderDefinition(device, options) {
        var shaderPassInfo = ShaderPass.get(device).getByIndex(options.litOptions.pass);
        var isForwardPass = shaderPassInfo.isForward;
        var litShader = new LitShader(device, options.litOptions, SHADERLANGUAGE_GLSL);
        // generate vertex shader
        this.createVertexShader(litShader, options);
        // handle fragment shader
        var textureMapping = {};
        options.litOptions.fresnelModel = options.litOptions.fresnelModel === 0 ? FRESNEL_SCHLICK : options.litOptions.fresnelModel;
        // fragment defines
        var fDefines = litShader.fDefines;
        this.prepareFragmentDefines(options, fDefines, shaderPassInfo);
        var lightingUv = '';
        if (isForwardPass) {
            // parallax
            if (options.heightMap) {
                this._addMapDefines(fDefines, 'height', 'parallaxPS', options, litShader.chunks, textureMapping);
            }
            // opacity
            if (options.litOptions.blendType !== BLEND_NONE || options.litOptions.alphaTest || options.litOptions.alphaToCoverage || options.litOptions.opacityDither !== DITHER_NONE) {
                this._addMapDefines(fDefines, 'opacity', 'opacityPS', options, litShader.chunks, textureMapping);
            }
            // normal
            if (litShader.needsNormal) {
                if (options.normalMap || options.clearCoatNormalMap) {
                    if (!options.litOptions.hasTangents) {
                        // TODO: generalize to support each normalmap input (normalMap, normalDetailMap, clearCoatNormalMap) independently
                        var baseName = options.normalMap ? 'normalMap' : 'clearCoatNormalMap';
                        lightingUv = this._getUvSourceExpression("" + baseName + "Transform", "" + baseName + "Uv", options);
                    }
                }
                this._addMapDefines(fDefines, 'normalDetail', 'normalMapPS', options, litShader.chunks, textureMapping, options.normalDetailPackedNormal ? 'xy' : 'xyz');
                this._addMapDefines(fDefines, 'normal', 'normalMapPS', options, litShader.chunks, textureMapping, options.packedNormal ? 'xy' : 'xyz');
            }
            // diffuse
            if (options.diffuseDetail) {
                this._addMapDefines(fDefines, 'diffuseDetail', 'diffusePS', options, litShader.chunks, textureMapping, options.diffuseDetailEncoding);
            }
            this._addMapDefines(fDefines, 'diffuse', 'diffusePS', options, litShader.chunks, textureMapping, options.diffuseEncoding);
            // refraction
            if (options.litOptions.useRefraction) {
                this._addMapDefines(fDefines, 'refraction', 'transmissionPS', options, litShader.chunks, textureMapping);
                this._addMapDefines(fDefines, 'thickness', 'thicknessPS', options, litShader.chunks, textureMapping);
            }
            // iridescence
            if (options.litOptions.useIridescence) {
                this._addMapDefines(fDefines, 'iridescence', 'iridescencePS', options, litShader.chunks, textureMapping);
                this._addMapDefines(fDefines, 'iridescenceThickness', 'iridescenceThicknessPS', options, litShader.chunks, textureMapping);
            }
            // specularity & glossiness
            if (litShader.lighting && options.litOptions.useSpecular || litShader.reflections) {
                if (options.litOptions.useSheen) {
                    this._addMapDefines(fDefines, 'sheen', 'sheenPS', options, litShader.chunks, textureMapping, options.sheenEncoding);
                    this._addMapDefines(fDefines, 'sheenGloss', 'sheenGlossPS', options, litShader.chunks, textureMapping);
                }
                if (options.litOptions.useMetalness) {
                    this._addMapDefines(fDefines, 'metalness', 'metalnessPS', options, litShader.chunks, textureMapping);
                    this._addMapDefines(fDefines, 'ior', 'iorPS', options, litShader.chunks, textureMapping);
                }
                if (options.litOptions.useSpecularityFactor) {
                    this._addMapDefines(fDefines, 'specularityFactor', 'specularityFactorPS', options, litShader.chunks, textureMapping);
                }
                if (options.useSpecularColor) {
                    this._addMapDefines(fDefines, 'specular', 'specularPS', options, litShader.chunks, textureMapping, options.specularEncoding);
                }
                this._addMapDefines(fDefines, 'gloss', 'glossPS', options, litShader.chunks, textureMapping);
            }
            // ao
            if (options.aoDetail) {
                this._addMapDefines(fDefines, 'aoDetail', 'aoPS', options, litShader.chunks, textureMapping);
            }
            if (options.aoMap || options.aoVertexColor || options.useAO) {
                this._addMapDefines(fDefines, 'ao', 'aoPS', options, litShader.chunks, textureMapping);
            }
            // emission
            this._addMapDefines(fDefines, 'emissive', 'emissivePS', options, litShader.chunks, textureMapping, options.emissiveEncoding);
            // clearcoat
            if (options.litOptions.useClearCoat) {
                this._addMapDefines(fDefines, 'clearCoat', 'clearCoatPS', options, litShader.chunks, textureMapping);
                this._addMapDefines(fDefines, 'clearCoatGloss', 'clearCoatGlossPS', options, litShader.chunks, textureMapping);
                this._addMapDefines(fDefines, 'clearCoatNormal', 'clearCoatNormalPS', options, litShader.chunks, textureMapping, options.clearCoatPackedNormal ? 'xy' : 'xyz');
            }
            // lightmap
            if (options.lightMap || options.lightVertexColor) {
                this._addMapDefines(fDefines, 'light', 'lightmapPS', options, litShader.chunks, textureMapping, options.lightMapEncoding);
            }
        } else {
            // all other passes require only opacity
            var opacityShadowDither = options.litOptions.opacityShadowDither;
            if (options.litOptions.alphaTest || opacityShadowDither) {
                this._addMapDefines(fDefines, 'opacity', 'opacityPS', options, litShader.chunks, textureMapping);
            }
        }
        // generate fragment shader - supply the front end declaration and code to the lit shader, which will
        // use the outputs from this standard shader generator as an input to the lit shader generator
        litShader.generateFragmentShader(litShader.chunks.stdDeclarationPS, litShader.chunks.stdFrontEndPS, lightingUv);
        var includes = new Map(Object.entries(_extends({}, Object.getPrototypeOf(litShader.chunks), litShader.chunks, options.litOptions.chunks)));
        var vDefines = litShader.vDefines;
        options.defines.forEach((value, key)=>vDefines.set(key, value));
        options.defines.forEach((value, key)=>fDefines.set(key, value));
        var definition = ShaderUtils.createDefinition(device, {
            name: 'StandardShader',
            attributes: litShader.attributes,
            vertexCode: litShader.vshader,
            fragmentCode: litShader.fshader,
            vertexIncludes: includes,
            fragmentIncludes: includes,
            fragmentDefines: fDefines,
            vertexDefines: vDefines
        });
        if (litShader.shaderPassInfo.isForward) {
            definition.tag = SHADERTAG_MATERIAL;
        }
        return definition;
    }
    constructor(...args){
        super(...args), // Shared Standard Material option structures
        this.optionsContext = new StandardMaterialOptions(), this.optionsContextMin = new StandardMaterialOptions();
    }
}
var standard = new ShaderGeneratorStandard();

export { _matTex2D, standard };
